use std::{borrow::Cow, collections::HashMap};

use db::folder::FolderWithoutParentId;

use crate::prelude::*;

#[derive(sqlx::Type, Debug, Serialize, Deserialize)]
#[serde(rename_all = "snake_case")]
#[sqlx(type_name = "permission")]
#[sqlx(rename_all = "lowercase")]
pub enum PermissionRaw {
    Read = 1,
    Write,
    Manage,
}

#[derive(Clone, Copy, Debug, PartialEq, Eq, PartialOrd, Ord, Default)]
pub enum PermissionType {
    #[default]
    NoPermission = 1,
    Read,
    Write,
    Manage,
}

impl From<Option<PermissionRaw>> for PermissionType {
    fn from(value: Option<PermissionRaw>) -> PermissionType {
        use PermissionRaw as PR;
        match value {
            Some(PR::Read) => PermissionType::Read,
            Some(PR::Write) => PermissionType::Write,
            Some(PR::Manage) => PermissionType::Manage,
            None => PermissionType::NoPermission,
        }
    }
}

impl PermissionType {
    pub fn can_read(self) -> bool {
        self >= PermissionType::Read
    }

    fn can_read_guard(self) -> GeneralResult<()> {
        self.can_read().then_some(()).item_not_found()?;
        Ok(())
    }

    fn can_write_guard(self) -> GeneralResult<()> {
        self.can_read_guard()?;
        if self < PermissionType::Write {
            return Err(GeneralError::message(
                StatusCode::FORBIDDEN,
                "Cannot write to the folder",
            ));
        }
        Ok(())
    }

    fn can_manage_guard(self) -> GeneralResult<()> {
        self.can_read_guard()?;
        if self < PermissionType::Manage {
            return Err(GeneralError::message(
                StatusCode::FORBIDDEN,
                "Cannot manage the folder",
            ));
        }
        Ok(())
    }
}

pub trait PermissionExt {
    fn can_read_guard(self) -> GeneralResult<()>;
    fn can_write_guard(self) -> GeneralResult<()>;
    fn can_manage_guard(self) -> GeneralResult<()>;
}

fn permissions_error(error: sqlx::Error) -> GeneralError {
    GeneralError {
        status_code: StatusCode::INTERNAL_SERVER_ERROR,
        message: Cow::Borrowed("Error getting permissions"),
        error: Some(error.into()),
    }
}

fn apply_guard(
    result: sqlx::Result<PermissionType>,
    func: impl FnOnce(PermissionType) -> GeneralResult<()>,
) -> GeneralResult<()> {
    result.map_err(permissions_error).and_then(func)
}

impl PermissionExt for sqlx::Result<PermissionType> {
    fn can_read_guard(self) -> GeneralResult<()> {
        apply_guard(self, PermissionType::can_read_guard)
    }

    fn can_write_guard(self) -> GeneralResult<()> {
        apply_guard(self, PermissionType::can_write_guard)
    }

    fn can_manage_guard(self) -> GeneralResult<()> {
        apply_guard(self, PermissionType::can_manage_guard)
    }
}

pub async fn insert(
    user_id: i32,
    folder_id: Uuid,
    permission_type: PermissionRaw,
    pool: &Pool,
) -> sqlx::Result<()> {
    sqlx::query_file!(
        "sql/insert_permission.sql",
        folder_id,
        user_id,
        permission_type as PermissionRaw
    )
    .execute(pool)
    .await
    .map(|_| ())
}

pub async fn get_all_for_folder(
    folder_id: Uuid,
    pool: &Pool,
) -> sqlx::Result<HashMap<i32, PermissionRaw>> {
    sqlx::query_file!("sql/get_all_permissions_for_folder.sql", folder_id)
        .fetch(pool)
        .map_ok(|record| (record.user_id, record.permission_type))
        .try_collect()
        .await
}

pub async fn delete_for_folder(folder_id: Uuid, user_id: i32, pool: &Pool) -> sqlx::Result<()> {
    sqlx::query_file!("sql/delete_permissions.sql", folder_id, user_id)
        .execute(pool)
        .await
        .map(|_| ())
}

pub async fn get_top_level_permitted_folders(
    user_id: i32,
    pool: &Pool,
) -> sqlx::Result<Vec<FolderWithoutParentId>> {
    sqlx::query_file_as!(
        FolderWithoutParentId,
        "sql/get_top_level_folder.sql",
        user_id
    )
    .fetch(pool)
    .try_collect()
    .await
}
