use std::{
    collections::{HashMap, HashSet},
    fmt::Write as _,
};

use axum::extract::multipart::{self, Multipart};

use crate::prelude::*;

#[derive(Debug, Deserialize)]
pub struct Params {
    parent_folder: Uuid,
}

#[derive(Serialize, Debug, Default)]
pub struct Response {
    success: HashMap<Box<str>, Uuid>,
    error: HashMap<Box<str>, &'static str>,
}

fn validate_name(name: &str, existing_names: &HashSet<String>) -> Result<(), &'static str> {
    if name.len() > 255 {
        return Err("Name too long");
    }
    if existing_names.contains(name) {
        return Err("Item with that name already exists");
    }
    Ok(())
}

async fn create_file(
    storage: &crate::FileStorage,
    file_name: &str,
    field: &mut multipart::Field<'_>,
    parent_folder: Uuid,
    pool: &Pool,
) -> anyhow::Result<Uuid> {
    let (file_id, file) = storage.create().await?;
    let result = async {
        let (hash, size) = crate::FileStorage::write_to_file(file, field).await?;
        db::file::insert(file_id, parent_folder, file_name, size, hash, pool).await?;
        anyhow::Result::Ok(())
    }
    .await;
    match result {
        Ok(()) => Ok(file_id),
        Err(err) => {
            let _ = storage.delete(file_id).await;
            Err(err)
        }
    }
}

async fn parse_field(
    field: &mut multipart::Field<'_>,
    name: &str,
    storage: &crate::FileStorage,
    parent_folder: Uuid,
    pool: &Pool,
    existing_names: &HashSet<String>,
) -> Result<Uuid, &'static str> {
    validate_name(name, existing_names)?;
    create_file(storage, name, field, parent_folder, pool)
        .await
        .map_err(|err| {
            tracing::warn!(%err, "Error creating the file");
            "Error creating the file"
        })
}

pub async fn upload(
    Query(params): Query<Params>,
    State(state): State<AppState>,
    claims: Claims,
    mut multi: Multipart,
) -> GeneralResult<Json<Response>> {
    db::folder::get_permissions(params.parent_folder, claims.user_id, &state.pool)
        .await
        .can_write_guard()?;

    let existing_names: HashSet<String> = db::folder::get_names(params.parent_folder, &state.pool)
        .try_collect()
        .await
        .handle_internal("Error getting existing names")?;

    let mut response = Response::default();
    while let Ok(Some(mut field)) = multi.next_field().await {
        let Some(file_name) = field.file_name().map(Box::<str>::from) else {
            continue;
        };

        let parse_result = parse_field(
            &mut field,
            &file_name,
            &state.storage,
            params.parent_folder,
            &state.pool,
            &existing_names,
        )
        .await;

        match parse_result {
            Ok(uuid) => {
                response.success.insert(file_name, uuid);
            }
            Err(err) => {
                response.error.insert(file_name, err);
            }
        }
    }

    if !response.success.is_empty() {
        return Ok(Json(response));
    }

    if response.error.is_empty() {
        return Err(GeneralError::message(
            StatusCode::BAD_REQUEST,
            "No files sent",
        ));
    }

    let mut message = "No file successfully uploaded:".to_owned();
    for (key, val) in response.error {
        write!(message, "\n{key}: {val}").unwrap();
    }

    Err(GeneralError::message(StatusCode::BAD_REQUEST, message))
}
