use std::borrow::Cow;

use axum::{http::StatusCode, response::IntoResponse};

type BoxError = Box<dyn std::error::Error>;

/// Common error type for the project
pub struct GeneralError {
    /// Response status code
    pub status_code: StatusCode,
    /// Message to send to the user
    pub message: Cow<'static, str>,
    /// Error to log
    pub error: Option<BoxError>,
}

impl GeneralError {
    pub fn message(status_code: StatusCode, message: impl Into<Cow<'static, str>>) -> Self {
        Self {
            status_code,
            message: message.into(),
            error: None,
        }
    }

    pub const fn const_message(status_code: StatusCode, message: &'static str) -> Self {
        Self {
            status_code,
            message: Cow::Borrowed(message),
            error: None,
        }
    }
}

impl IntoResponse for GeneralError {
    fn into_response(self) -> axum::response::Response {
        if let Some(err) = self.error {
            tracing::error!(err, message = %self.message, status_code = ?self.status_code);
        }
        (self.status_code, self.message).into_response()
    }
}

impl From<validator::ValidationErrors> for GeneralError {
    fn from(value: validator::ValidationErrors) -> Self {
        GeneralError::message(StatusCode::BAD_REQUEST, value.to_string())
    }
}

pub type GeneralResult<T> = Result<T, GeneralError>;

pub trait ErrorHandlingExt<T>
where
    Self: Sized,
{
    fn handle(
        self,
        status_code: StatusCode,
        message: impl Into<Cow<'static, str>>,
    ) -> GeneralResult<T>;

    fn handle_internal(self, message: impl Into<Cow<'static, str>>) -> GeneralResult<T> {
        self.handle(StatusCode::INTERNAL_SERVER_ERROR, message)
    }
}

impl<T, E: Into<BoxError>> ErrorHandlingExt<T> for Result<T, E> {
    fn handle(
        self,
        status_code: StatusCode,
        message: impl Into<Cow<'static, str>>,
    ) -> GeneralResult<T> {
        self.map_err(|err| GeneralError {
            status_code,
            message: message.into(),
            error: Some(err.into()),
        })
    }
}

impl<T> ErrorHandlingExt<T> for Option<T> {
    fn handle(
        self,
        status_code: StatusCode,
        message: impl Into<Cow<'static, str>>,
    ) -> GeneralResult<T> {
        self.ok_or_else(|| GeneralError {
            status_code,
            message: message.into(),
            error: None,
        })
    }
}

pub trait ItemNotFoundExt<T> {
    fn item_not_found(self) -> Result<T, GeneralError>;
}

impl<T> ItemNotFoundExt<T> for Option<T> {
    fn item_not_found(self) -> GeneralResult<T> {
        self.handle(StatusCode::NOT_FOUND, "Item not found")
    }
}
