mod auth;
mod db;
mod endpoints;
mod errors;
mod file_storage;
mod prelude;
mod util;

use file_storage::FileStorage;

type Pool = sqlx::postgres::PgPool;

#[derive(Clone, axum::extract::FromRef)]
struct AppState {
    pool: Pool,
    storage: FileStorage,
}

async fn create_test_users(pool: &Pool) -> anyhow::Result<()> {
    let count = sqlx::query!("SELECT count(user_id) FROM users")
        .fetch_one(pool)
        .await?
        .count
        .unwrap_or(0);
    if count > 0 {
        return Ok(());
    }
    let hash1 = auth::HashedBytes::hash_bytes(b"Password1").as_bytes();
    let hash2 = auth::HashedBytes::hash_bytes(b"Password2").as_bytes();

    tokio::try_join!(
        db::users::create_user("Test1", "test1@example.com", &hash1, pool),
        db::users::create_user("Test2", "test2@example.com", &hash2, pool)
    )?;

    Ok(())
}

fn init_tracing() {
    use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt};

    let mut err = None;

    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env().unwrap_or_else(|inner_err| {
                err = Some(inner_err);
                "debug,sqlx=info,axum::rejection=trace".parse().unwrap()
            }),
        )
        .with(tracing_subscriber::fmt::layer())
        .init();

    if let Some(err) = err {
        tracing::info!(
            %err,
            "Error constructing EnvFilter, falling back to using the default"
        );
    }
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    use std::{env, net::Ipv4Addr};

    if env::var("RUST_BACKTRACE").is_err() {
        env::set_var("RUST_BACKTRACE", "1");
    }
    let _ = dotenvy::dotenv();

    init_tracing();

    auth::force_init_keys();

    let pool = match env::var("DATABASE_URL") {
        Ok(url) => Pool::connect(&url).await?,
        Err(err) => anyhow::bail!("Error getting database url: {err}"),
    };
    sqlx::migrate!().run(&pool).await?;
    if let Ok("1") = env::var("DEVELOPMENT").as_deref().map(str::trim_ascii) {
        create_test_users(&pool).await?;
    }

    let state = AppState {
        pool,
        storage: FileStorage::new()?,
    };

    let router = app(state);
    let addr = (Ipv4Addr::UNSPECIFIED, 3000);
    let listener = tokio::net::TcpListener::bind(addr).await?;

    axum::serve(listener, router)
        .with_graceful_shutdown(shutdown_signal())
        .await?;

    Ok(())
}

async fn shutdown_signal() {
    use tokio::signal;

    let ctrl_c = async {
        signal::ctrl_c()
            .await
            .expect("failed to install Ctrl+C handler");
    };

    #[cfg(unix)]
    {
        let terminate = async {
            signal::unix::signal(signal::unix::SignalKind::terminate())
                .expect("failed to install signal handler")
                .recv()
                .await;
        };
        tokio::select! {
            () = ctrl_c => {}
            () = terminate => {}
        }
    }

    #[cfg(not(unix))]
    ctrl_c.await;
}

fn app(state: AppState) -> axum::Router {
    use axum::{
        extract::DefaultBodyLimit,
        handler::Handler as _,
        http::header,
        routing::{get, post},
    };
    use endpoints::{
        file, folder,
        permissions::{self, get_top_level::get_top_level},
        users,
    };
    use tower_http::{
        timeout::TimeoutLayer,
        trace::{MakeSpan, TraceLayer},
        ServiceBuilderExt as _,
    };

    #[derive(Clone, Copy)]
    struct SpanMaker;

    impl<B> MakeSpan<B> for SpanMaker {
        fn make_span(&mut self, request: &axum::http::Request<B>) -> tracing::Span {
            tracing::debug_span!(
                "request",
                method = %request.method(),
                uri = %request.uri(),
                version = ?request.version(),
                headers = ?request.headers(),
                request_id = %uuid::Uuid::now_v7()
            )
        }
    }

    const TEN_GIBIBYTES: usize = 10 * 1024 * 1024 * 1024;
    let body_limit = DefaultBodyLimit::max(TEN_GIBIBYTES);

    let timeout = TimeoutLayer::new(std::time::Duration::from_secs(10));

    let common_middleware = tower::ServiceBuilder::new()
        .sensitive_headers([header::AUTHORIZATION, header::COOKIE])
        .layer(TraceLayer::new_for_http().make_span_with(SpanMaker))
        .compression();

    let file_router = axum::Router::new().route(
        "/",
        get(file::download::download)
            .post(file::upload::upload.layer(body_limit.clone()))
            .delete(file::delete::delete.layer(timeout))
            .patch(file::modify::modify.layer(body_limit.clone())),
    );

    let general_router = axum::Router::new()
        .route(
            "/folders",
            get(folder::list::list)
                .post(folder::create::create)
                .delete(folder::delete::delete),
        )
        .route("/folders/structure", get(folder::get_structure::structure))
        .route(
            "/permissions",
            get(permissions::get::get)
                .post(permissions::set::set)
                .delete(permissions::delete::delete),
        )
        .route(
            "/permissions/get_top_level_permitted_folders",
            get(get_top_level),
        )
        .route(
            "/users",
            get(users::get::get)
                .delete(users::delete::delete)
                .put(users::put::put),
        )
        .route("/users/current", get(users::get::current))
        .route("/users/search", get(users::search::search))
        .route("/users/register", post(users::register::register))
        .route("/users/authorize", post(users::login::login))
        .layer(timeout);

    axum::Router::new()
        .nest("/files", file_router)
        .nest("/", general_router)
        .layer(common_middleware)
        .with_state(state)
}
